import React from "react";
import './index.css'
import { getTop, getLeft, clearEventBubble, getScroll, isInPath, getElementsClass, isEqual } from './utils'

interface IProps{
    wrapperScroll?: any,
    itemClass: string,
    activeClass?: string,
    isDataChange: any,
    activePosition?: any,
    onSelected: (pos: any, item: any, data: any) => void,
    onSingleSelected?: (pos: any, item: any, data: any) => void,
}
interface IState{
    isMouseDown: boolean,
    allItems: any[],
    selectedItems: any[],
    selectedPostions: any[],
    selectedPostionsHold: any[],
    selectEle: any,
    startX: number,
    startY: number
}

/**
 * wrapperScroll?: any, // 滚动单位'ID';
 * 
 * itemClass: string,  // 列表框通用携带 class
 * 
 * activeClass?: string, // 列表框 选中class
 * 
 * isDataChange: any,  // 渲染的数组，检测数据更改
 * 
 * activePosition?: any,  // 主动选中数据
 * 
 * onSelected: (pos: any, item: any, data: any) => void, // 拖动
 * 
 * onSingleSelected?: (pos: any, item: any, data: any) => void, // 单击
 * 
 * 注：itemClass 与 children 循环box className必须携带一致，同box必须绑定data-position，data-position导出选中的唯一识别
 * 注：带问号为可选项，否则都是必填项。
 */
export class MouseMultiples extends React.Component<IProps, IState> {
    private constructor(props: IProps) {
        super(props);
        this.state = {
            isMouseDown: false,
            allItems: [],
            selectedItems: [],
            selectedPostions: [],
            selectedPostionsHold: [],
            selectEle: {
                width: 0,
                height: 0,
                top: 0,
                left: 0,
                display: 'none'
            },
            startX: 0,
            startY: 0
        };
    }

    public onKeyDown = (e: { ctrlKey: any; keyCode: number; }) => {
        if(e.ctrlKey && e.keyCode === 65){
            this.stopDefault(e)
            const allActive: string | any[] = []
            const { allItems  } = this.state
            allItems.forEach(i => {
                allActive.push(i.dataset['position'])
            })
            this.handldImportActiveClass(allActive)
        }
    }

    public stopDefault(e: { ctrlKey?: any; keyCode: any; preventDefault?: any; }) {
        if (e?.preventDefault) {
            e.preventDefault();
        } else {
            e.keyCode = 0;
        }
    }

    public componentDidMount () {
        this.hanldAllItems()
        document.addEventListener("keydown", this.onKeyDown)
    }

    public componentWillUnmount () {
        this.setState({ allItems: [], selectedItems: [] })
        this.setState = () => { }
        document.removeEventListener("keydown", this.onKeyDown)
    }

    public componentDidUpdate(prevProps:any) {
        // 典型用法（不要忘记比较 props）：
        if (!isEqual(this.props.isDataChange, prevProps.isDataChange)) {
            this.upSelectedPostionsHold()
        }
        if (prevProps.activePosition !== this.props.activePosition) {
            this.handldImportActiveClass(this.props.activePosition)
        }
    }

    public upSelectedPostionsHold (){
        this.setState({ selectedPostionsHold: [] })
        this.hanldAllItems()
    }

    public hanldAllItems (){
        const { itemClass } = this.props
        const allItems = getElementsClass(itemClass)
        this.setState({ allItems })
    }

    public handldImportActiveClass(activeId: string | any[]){
        const { activeClass, onSelected } = this.props
        const { selectedItems, selectedPostions, allItems  } = this.state
        selectedItems.length = 0
        selectedPostions.length = 0
        allItems.forEach(i => {
            if(activeId.indexOf(i.dataset['position']) !== -1){
                selectedItems.push(i)
                selectedPostions.push(i.dataset['position'])
                i.classList.add(activeClass || 'selection_item_active')
            }else{
                i.classList.remove(activeClass || 'selection_item_active')
            }
        })
        onSelected?.(selectedPostions, selectedItems, allItems)
        const selectedPostionsHold = selectedPostions.slice()
        this.setState({ selectedItems, selectedPostions: [], selectedPostionsHold })
    }

    public addActiveClass = (target: { classList: { add: (arg0: string) => void; }; }) => {
        const { activeClass } = this.props
        target.classList.add(activeClass || 'selection_item_active')
    }

    public removeActiveClass = (target: { classList: { remove: (arg0: string) => void; }; }) => {
        const { activeClass } = this.props
        target.classList.remove(activeClass || 'selection_item_active')
    }

    public creatSelectEle = (evt: { currentTarget?: any; clientX?: any; clientY?: any; }) => {
        const { selectEle, startX, startY } = this.state
        const { wrapperScroll } = this.props
        const top = getTop(evt.currentTarget)
        const left = getLeft(evt.currentTarget)
        const { scrollTop, scrollLeft } = getScroll(wrapperScroll)
    
        const { clientX, clientY } = evt
        const _x = clientX - left + scrollLeft
        const _y = clientY - top + scrollTop
    
        selectEle.left = _x > startX ? startX - 1 : _x + 1
        selectEle.top = _y > startY ? startY - 1 : _y + 1
        selectEle.width = Math.abs(_x - startX)
        selectEle.height = Math.abs(_y - startY)
        selectEle.display = 'block'

        return selectEle
    }

    public down = (e: { currentTarget?: any; clientX?: any; clientY?: any; }) => {
        // clearEventBubble(e)
        const { selectEle } = this.state
        const { wrapperScroll } = this.props
    
        const { clientX, clientY } = e
    
        const top = getTop(e.currentTarget)
        const left = getLeft(e.currentTarget)
        const { scrollTop, scrollLeft } = getScroll(wrapperScroll)
    
        const startX = clientX - left + scrollLeft
        const startY = clientY - top + scrollTop
    
        selectEle.left = startX
        selectEle.top = startY
    
        this.setState({ isMouseDown: true, startX, startY, selectEle })
    }

    public isMove = (evt: { currentTarget?: any; clientX?: any; clientY?: any; }) => {
        let flag = false
        const { startX, startY } = this.state
        const { wrapperScroll } = this.props

        const top = getTop(evt.currentTarget)
        const left = getLeft(evt.currentTarget)
        const { clientX, clientY } = evt
        const { scrollTop, scrollLeft } = getScroll(wrapperScroll)

        const moveX = clientX - left + scrollLeft
        const moveY = clientY - top + scrollTop
        if(Math.abs(moveX - startX) < 5 || Math.abs(moveY - startY) < 5){
            flag = true
        }
        return flag
    }

    public move = (e: { currentTarget: any; clientX: any; clientY: any; }) => {
        clearEventBubble(e)
        const { isMouseDown, selectedItems, selectedPostions, allItems, selectedPostionsHold } = this.state
    
        if (!isMouseDown) return
        if (this.isMove(e)) return
    
        const selectEle = this.creatSelectEle(e)
        selectedPostionsHold.length = 0
        selectedPostions.length = 0
        selectedItems.length = 0
        
        allItems.forEach(item => {
          this.removeActiveClass(item)
    
          if (isInPath(item, selectEle)) {
            this.addActiveClass(item)
    
            selectedPostions.push(item.dataset['position'])
            selectedItems.push(item)
          }
        })
        this.setState({ selectEle, selectedItems, selectedPostions, selectedPostionsHold })
    }

    public up = (e: { [x: string]: any; currentTarget?: any; clientX?: any; clientY?: any; }) => {
        clearEventBubble(e)
        let { allItems, selectedPostionsHold, selectedPostions, selectedItems, isMouseDown } = this.state
        const { onSelected } = this.props
        if (!isMouseDown) return
        const selectEle = this.creatSelectEle(e)
        const isCtrlKey = e['ctrlKey']
        if(isCtrlKey){
            selectedPostions.length = 0
        }
        if (selectedPostions.length === 0) {
            let noItem = true
            if(isCtrlKey){
                allItems.forEach(item => {
                    if (isInPath(item, selectEle)) {
                        noItem = false
                        const itemPosition = item.dataset['position']
                        if(selectedPostionsHold.indexOf(itemPosition) !== -1){
                            this.removeActiveClass(item)
                            selectedPostions = selectedPostionsHold.filter(i => i !== itemPosition)
                            selectedItems = selectedItems.filter(i => i.dataset['position'] !== itemPosition )
                        } else {
                            this.addActiveClass(item)
                            selectedPostions = [...selectedPostionsHold, itemPosition]
                            selectedItems.push(item)
                        }
                    }
                })
                if(noItem){
                    selectedPostions = [...selectedPostionsHold]
                }
            }else{
                allItems.forEach(item => {
                    if (isInPath(item, selectEle)) {
                        noItem = false
                    }
                })
                if(noItem){
                    selectedPostions = [...selectedPostionsHold]
                }else{
                    allItems.forEach(item => {
                        this.removeActiveClass(item)
                        if (isInPath(item, selectEle)) {
                            noItem = false
                            this.addActiveClass(item)
                            selectedPostions.push(item.dataset['position'])
                            selectedItems.push(item)
                        }
                    })
                }
            }
            !noItem && onSelected?.(selectedPostions, selectedItems, allItems)
        } else {
            onSelected?.(selectedPostions, selectedItems, allItems)
        }

        selectEle.left = 0
        selectEle.top = 0
        selectEle.width = 0
        selectEle.height = 0
        selectEle.display = 'none'
        let holds = selectedPostions.slice()
        this.setState({ isMouseDown: false, selectEle, selectedPostionsHold:holds, selectedPostions: [] })
    }
    
    public leave = (e: any) => {
        const { isMouseDown } = this.state
        if (!isMouseDown) return
        let evt = {...e}
        setTimeout(() => this.up(evt), 300)
    }

    public render () {
        const { selectEle } = this.state
        const { children } = this.props
        return (
            <div 
            className='mouse_selection'
            onMouseDown={this.down}
            onMouseMove={this.move}
            onMouseUp={this.up}
            onMouseLeave={this.leave}
            >
                <div className='selection_element' style={{ ...selectEle }} />
                {children}
            </div>
        )
    }
}

export default MouseMultiples